"""Package teneva, module grid: functions to operate with grids.

This module contains a set of functions for creating and transforming
multidimensional grids for discretization on uniform and Chebyshev grids.

"""
import itertools
import numpy as np


def grid_flat(n):
    """Compute the multi-indices for the full multidimensional grid.

    Args:
        n (int, float, list, np.ndarray): number of grid points for each
            dimension (list or np.ndarray of length d, where d is a number
            of dimensions). It may be also a number, then the 1D grid will be
            returned.

    Returns:
        np.ndarray: multi-indices for the full (flatten) grid (it is 2D array of
        the shape [d, n^d]).

    """
    if isinstance(n, (int, float, np.int32, np.float32, np.int64, np.float64)):
        return np.arange(int(n))

    d = len(n)
    I = [np.arange(k).reshape(1, -1) for k in n]
    I = np.meshgrid(*I, indexing='ij')
    I = np.array(I, dtype=int).reshape((d, -1), order='F').T
    return I


def grid_prep_opt(opt, d=None, kind=float, reps=None):
    """Helper function that prepare grid parameter.

    Args:
        opt (int, float, list, np.ndarray): grid parameter for each dimension.
        d (int): the dimension of the grid. If should be set if opt is scalar.
        kind (class 'int', class 'float'): data type for option (int or float).
        reps (int): optional number of repetitions for option.

    Returns:
        np.ndarray: prepared option. It is 1D np.ndarray of length d if reps
        is None, or np.ndarray of the shape [reps, d] (values repeated along
        the 1th axis) otherwise.

    """
    if opt is None:
        return None

    if isinstance(opt, (int, float)):
        if d is None or d <= 0:
            raise ValueError('Invalid grid option')
        opt = np.ones(d, dtype=kind) * kind(opt)
    opt = np.asanyarray(opt, dtype=kind)

    if reps is not None:
        opt = np.repeat(opt.reshape((1, -1)), reps, axis=0)

    return opt


def grid_prep_opts(a=None, b=None, n=None, d=None, reps=None):
    """Helper function that prepare grid parameters a, b, n.

    Args:
        a (float, list, np.ndarray): grid lower bounds for each dimension (list
            or np.ndarray of length d or float).
        b (float, list, np.ndarray): grid upper bounds for each dimension (list
            or np.ndarray of length d or float).
        n (int, float, list, np.ndarray): grid size for each dimension (list or
            np.ndarray of length d or int/float).
        d (int): the dimension of the grid. If all a, b, n are numbers,
            then it should be set, otherwise it is optional (d will be
            recoverd fromv the given a / b / n).
        reps (int): optional number of repetitions for a, b, n.

    Returns:
        (np.ndarray, np.ndarray, np.ndarray): grid lower bounds a, grid upper
        bounds b and grid size n for each dimension. All opts will be 1D
        arrays of length d if reps is None, or np.ndarray of the shape
        [reps, d] (values repeated along the 1th axis) otherwise.

    Note:
        In case of a mismatch in the size of the arrays a / b / n or the
        impossibility of restoring the dimension d, the function will
        generate the error (ValueError).

    """
    for item in [a, b, n]:
        if isinstance(item, (list, np.ndarray)):
            if d is None:
                d = len(item)
            elif d != len(item):
                raise ValueError('Invalid grid option')

    a = grid_prep_opt(a, d, float, reps)
    b = grid_prep_opt(b, d, float, reps)
    n = grid_prep_opt(n, d, int, reps)

    return a, b, n


def ind_qtt_to_tt(I_qtt, q):
    """Transform tensor multi-indices from QTT (long) to base TT (short) format.

    Args:
        I_qtt (list, np.ndarray): QTT multi-indices for the tensor in the form
            of array of the shape [samples, d*q], where samples is the number
            of samples, d is the dimension of the TT-tensor and q is a
            quantization value. For the case of only one sample, it may be 1D
            array or list of the length d*q.
        q (int): quantization value (TT-tensor mode size will be n=2^q).

    Returns:
        np.ndarray: TT multi-indices, which relates to the given QTT
        multi-indices in the form of array of the shape [samples, d]. If input
        I_qtt is 1D list or np.ndarray (the case of only one sample), then
        function will also return 1D np.ndarray of the length d.

    """
    I_qtt = grid_prep_opt(I_qtt, kind=int)

    if len(I_qtt.shape) == 1:
        is_many = False
        I_qtt = I_qtt.reshape(1, -1)
    else:
        is_many = True

    d = int(I_qtt.shape[1] / q)
    m = I_qtt.shape[0]
    n = [2]*q

    I = np.zeros((m, d), dtype=int)
    for i in range(d):
        I_qtt_curr = I_qtt[:, q*i:q*(i+1)].T
        I[:, i] = np.ravel_multi_index(I_qtt_curr, n, order='F')

    return I if is_many else I[0, :]


def ind_to_poi(I, a, b, n, kind='uni'):
    """Transform multi-indices (samples) into points of the spatial grid.

    Args:
        I (list, np.ndarray): multi-indices for the tensor in the form of array
            of the shape [samples, d], where samples is the number of samples
            and d is the dimension of the tensor. For the case of only one
            sample, it may be 1D array or list of length d.
        a (float, list, np.ndarray): grid lower bounds for each dimension (list
            or np.ndarray of length d). It may be also float, then the lower
            bounds for each dimension will be the same.
        b (float, list, np.ndarray): grid upper bounds for each dimension (list
            or np.ndarray of length d). It may be also float, then the upper
            bounds for each dimension will be the same.
        n (int, float, list, np.ndarray): tensor size for each dimension (list
            or np.ndarray of length d). It may be also int/float, then the
            size for each dimension will be the same.
        kind (str): the grid type, it may be "uni" (uniform grid) and "cheb"
            (Chebyshev grid). In case of the uniform grid, index 0 relates to
            the spatial point a and index n-1 relates to the spatial point
            b. In case of the Chebyshev grid, index 0 relates to the
            spatial point b and index n-1 relates to the spatial point a.

    Returns:
        np.ndarray: points of the grid in the form of array of the shape
        [samples, d]. If input I is 1D list or np.ndarray (the case of only
        one sample), then function will also return 1D np.ndarray of length d.

    Note:
        This function may be used for function approximation by low rank tensor
        methods (requested tensor indices should be transformed into spatial
        grid points and then used as an argument for the target function).

    """
    I = np.asanyarray(I)
    d = I.shape[-1]
    m = I.shape[0] if len(I.shape) > 1 else None
    a, b, n = grid_prep_opts(a, b, n, d, m)

    if kind == 'uni':
        X = I / (n - 1) * (b - a) + a
    elif kind == 'cheb':
        X = np.cos(np.pi * I / (n - 1)) * (b - a) / 2 + (b + a) / 2
    else:
        raise ValueError(f'Unknown grid type "{kind}"')

    return X


def ind_tt_to_qtt(I, n):
    """Transform tensor multi-indices from base TT (short) to QTT (long) format.

    Args:
        I (list, np.ndarray): TT multi-indices for the tensor in the form of
            array of the shape [samples, d], where samples is the number of
            samples and d is the dimension of the TT-tensor. For the case of
            only one sample, it may be 1D array or list of the length d.
        n (int): TT-tensor mode size. It should be like n=2^q, where q is a
            quantization value.

    Returns:
        np.ndarray: QTT multi-indices, which relates to the given TT
        multi-indices in the form of array of the shape [samples, d*q]. If input
        I is 1D list or np.ndarray (the case of only one sample), then
        function will also return 1D np.ndarray of the length d*q.

    """
    I = grid_prep_opt(I, kind=int)

    if len(I.shape) == 1:
        is_many = False
        I = I.reshape(1, -1)
    else:
        is_many = True

    d = int(I.shape[1])
    m = I.shape[0]
    q = int(np.log2(n))
    n_qtt = [2]*q

    if 2**q != n:
        raise ValueError('Invalid mode size (it should be a power of two)')

    I_qtt = np.zeros((m, d*q), dtype=int)
    for i in range(d):
        I_curr = I[:, i]
        I_qtt_curr = np.unravel_index(I_curr, n_qtt, order='F')
        I_qtt_curr = np.array(I_qtt_curr).T
        I_qtt[:, q*i:q*(i+1)] = I_qtt_curr

    return I_qtt if is_many else I_qtt[0, :]


def poi_scale(X, a, b, kind='uni'):
    """Scale points from [a, b] into unit interval.

    Args:
        X (list, np.ndarray): points of the spatial grid in the form of array of
            the shape [samples, d], where samples is the number of samples and
            d is the dimension of the tensor. For the case of only one sample,
            it may be 1D array or list of length d.
        a (float, list, np.ndarray): grid lower bounds for each dimension (list
            or np.ndarray of length d). It may be also float, then the lower
            bounds for each dimension will be the same.
        b (float, list, np.ndarray): grid upper bounds for each dimension (list
            or np.ndarray of length d). It may be also float, then the upper
            bounds for each dimension will be the same.
        kind (str): the grid type, it may be "uni" (uniform grid) and "cheb"
            (Chebyshev grid). It may be also a list of new scalar limits
            [a_new, b_new]; in this case the linear scaling to these limits
            will be performed.

    Returns:
        np.ndarray: scaled points of the spatial grid. It has the same shape as
        input array X. The interval will be [0, 1] in case of the uniform
        grid, and [-1, 1] in the case of the Chebyshev grid.

    """
    X = np.asanyarray(X, dtype=float)
    d = X.shape[-1]
    m = X.shape[0] if len(X.shape) > 1 else None

    a, b, _ = grid_prep_opts(a, b, None, d, m)

    if kind == 'uni':
        Xsc = (X - a) / (b - a)
        Xsc[Xsc < 0.] = 0.
        Xsc[Xsc > 1.] = 1.

    elif kind == 'cheb':
        Xsc = (X - (b + a) / 2) * (2 / (b - a))
        Xsc[Xsc < -1.] = -1.
        Xsc[Xsc > +1.] = +1.

    else:
        try:
            a_new, b_new = kind
        except ValueError:
            raise ValueError(f'Unknown grid type "{kind}"')

        Xsc = (X * (a_new - b_new) + a * b_new - b * a_new) / (a - b)
        Xsc[Xsc < a_new] = a_new
        Xsc[Xsc > b_new] = b_new

    return Xsc


def poi_to_ind(X, a, b, n, kind='uni'):
    """Transform points of the spatial grid (samples) into multi-indices.

    Args:
        X (list, np.ndarray): points of the spatial grid in the form of array of
            the shape [samples, d], where samples is the number of samples and
            d is the dimension of the tensor. For the case of only one
            sample, it may be 1D array or list of length d.
        a (float, list, np.ndarray): grid lower bounds for each dimension (list
            or np.ndarray of length d). It may be also float, then the lower
            bounds for each dimension will be the same.
        b (float, list, np.ndarray): grid upper bounds for each dimension (list
            or np.ndarray of length d). It may be also float, then the upper
            bounds for each dimension will be the same.
        n (int, float, list, np.ndarray): tensor size for each dimension (list
            or np.ndarray of length d). It may be also int/float, then the
            size for each dimension will be the same.
        kind (str): the grid type, it may be "uni" (uniform grid) and "cheb"
            (Chebyshev grid). In case of the uniform grid, index 0 relates to
            the spatial point a and index n-1 relates to the spatial point
            b. In case of the Chebyshev grid, index 0 relates to the
            spatial point b and index n-1 relates to the spatial point a.

    Returns:
        np.ndarray: multi-indices for the tensor in the form of array of the
        shape [samples, d]. If input X is 1D list or np.ndarray (the case of
        only one sample), then function will also return 1D np.ndarray of the
        length d.

    Note:
        Points that are outside the domain (a and b) will be transformed to
        the nearest grid indices (i.e., 0 or n-1).

    """
    Xsc = poi_scale(X, a, b, kind)
    d = Xsc.shape[-1]
    m = Xsc.shape[0] if len(Xsc.shape) > 1 else None
    n = grid_prep_opt(n, d, kind=int, reps=m)

    if kind == 'uni':
        I = Xsc * (n - 1)

    elif kind == 'cheb':
        I = np.arccos(Xsc) / np.pi * (n - 1)

    else:
        raise ValueError(f'Unknown grid type "{kind}"')

    I = np.rint(I)
    I = np.array(I, dtype=int)

    I[I < 0] = 0
    I[I > n-1] = n[I > n-1] - 1

    return I



"""Package teneva, module optima: estimate min and max value of the tensor.

This module contains the novel algorithm for computation of minimum and
maximum element of the given TT-tensor (function optima_tt).

"""
import numpy as np
import teneva


def optima_qtt(Y, k=100, e=1.E-12, r=100):
    """Find items which relate to min and max elements of the given TT-tensor.

    The provided TT-tensor Y is transformed into the QTT-format and then
    "optima_tt" method is applied to this QTT-tensor. Note that this method
    support only the tensors with constant mode size, which is a power of two,
    i.e., the shape should be [2^q, 2^q, ..., 2^q].

    Args:
        Y (list): d-dimensional TT-tensor of the shape [2^q, 2^q, ..., 2^q].
        k (int): number of selected items (candidates for the optimum) for each
            tensor mode.
        e (float): desired approximation accuracy for the QTT-tensor (> 0).
        r (int, float): maximum rank for the SVD decompositions while QTT-tensor
            construction (> 0).

    Returns:
        (np.ndarray, float, np.ndarray, float): multi-index (array of length d)
        which relates to minimum TT-tensor element; the related value of the
        tensor item (float); multi-index (array of length d) which relates to
        maximum TT-tensor element; the related value of the tensor item (float).
        I.e., the output looks like i_min, y_min, i_max, y_max.

    """
    n = teneva.shape(Y)

    for n_ in n[1:]:
        if n[0] != n_:
            msg = 'Invalid mode size (it should be equal for all modes)'
            raise ValueError(msg)

    n = n[0]
    q = int(np.log2(n))

    if 2**q != n:
        msg = 'Invalid mode size (it should be a power of two)'
        raise ValueError(msg)

    Z = teneva.tt_to_qtt(Y, e, r)

    i_min, y_min, i_max, y_max = optima_tt(Z, k)

    i_min = teneva.ind_qtt_to_tt(i_min, q)
    i_max = teneva.ind_qtt_to_tt(i_max, q)

    # We do this just in case to reduce the magnitude of numerical errors:
    y_min = teneva.get(Y, i_min)
    y_max = teneva.get(Y, i_max)

    return i_min, y_min, i_max, y_max


def optima_tt(Y, k=100):
    """Find items which relate to min and max elements of the given TT-tensor.

    Args:
        Y (list): d-dimensional TT-tensor.
        k (int): number of selected items (candidates for the optimum) for each
            tensor mode.

    Returns:
        (np.ndarray, float, np.ndarray, float): multi-index (array of length d)
        which relates to minimum TT-tensor element; the related value of the
        tensor item (float); multi-index (array of length d) which relates to
        maximum TT-tensor element; the related value of the tensor item (float).
        I.e., the output looks like i_min, y_min, i_max, y_max.

    Note:
        This function runs the "optima_tt_max" twice: first for the original
        tensor, and then for the tensor shifted by the value of the found
        maximum.

    """
    i1, y1 = optima_tt_max(Y, k)

    D = teneva.const(teneva.shape(Y), y1)
    Z = teneva.sub(Y, D)
    Z = teneva.mul(Z, Z)

    i2, _ = optima_tt_max(Z, k)
    y2 = teneva.get(Y, i2)

    if y2 > y1:
        return i1, y1, i2, y2
    else:
        return i2, y2, i1, y1


def optima_tt_beam(Y, k=100, l2r=True, ret_all=False, to_orth=True, p=None):
    """Find multi-index of the maximum modulo item in the given TT-tensor.

    Args:
        Y (list): d-dimensional TT-tensor.
        k (int): number of selected items (candidates for the optimum) for each
            tensor mode.
        l2r (bool): if flag is set, hen the TT-cores are passed from left to
            right (that is, from the first to the last TT-core). Otherwise, the
            TT-cores are passed from right to left.
        ret_all (bool): if flag is set, then all k multi-indices will be
            returned. Otherwise, only best found multi-index will be returned.

    Returns:
        np.ndarray: multi-index (array of length d) which relates to maximum
        modulo TT-tensor element if ret_all flag is not set. If ret_all flag
        is set, then it will be the set of k best multi-indices (array of the
        shape [k, d]).

    Note:
        This is an internal utility function. To find the optimum in the
        TT-tensor tensor, use the functions "optima_qtt", "optima_tt" or
        "optima_tt_max".

    """
    d = len(Y)
    if  to_orth:
        Z, p = teneva.orthogonalize(Y, 0 if l2r else len(Y)-1, use_stab=True)
    else:
        Z = Y
        if p is None:
            p = 1

    p0 = p / d # Scale factor (2^p0) for each TT-core

    G = Z[0 if l2r else -1]
    r1, n, r2 = G.shape

    I = teneva._range(n)
    Q = G.reshape(n, r2) if l2r else G.reshape(r1, n)

    Q *= 2**p0

    for G in (Z[1:] if l2r else Z[:-1][::-1]):
        r1, n, r2 = G.shape

        if l2r:
            Q = np.einsum('kr,riq->kiq', Q, G, optimize='optimal')
            Q = Q.reshape(-1, r2)
        else:
            Q = np.einsum('qir,rk->qik', G, Q, optimize='optimal')
            Q = Q.reshape(r1, -1)

        if l2r:
            I_l = np.kron(I, teneva._ones(n))
            I_r = np.kron(teneva._ones(I.shape[0]), teneva._range(n))
        else:
            I_l = np.kron(teneva._range(n), teneva._ones(I.shape[0]))
            I_r = np.kron(teneva._ones(n), I)
        I = np.hstack((I_l, I_r))

        q_max = np.max(np.abs(Q))
        norms = np.sum((Q/q_max)**2, axis=1 if l2r else 0)
        ind = np.argsort(norms)[:-(k+1):-1]

        I = I[ind, :]
        Q = Q[ind, :] if l2r else Q[:, ind]

        Q *= 2**p0

    return I if ret_all else I[0]


def optima_tt_max(Y, k=100):
    """Find the maximum modulo item in the given TT-tensor.

    Args:
        Y (list): d-dimensional TT-tensor.
        k (int): number of selected items (candidates for the optimum) for each
            tensor mode.

    Returns:
        (np.ndarray, float): multi-index (array of length d) which relates to
        maximum modulo TT-tensor element and the related value of the tensor
        item (float).

    Note:
        This function runs the "optima_tt_beam" first from left to right, then
        from right to left, and returns the best result.

    """
    i_max_list = [
        optima_tt_beam(Y, k, l2r=True),
        optima_tt_beam(Y, k, l2r=False)]
    y_max_list = [teneva.get(Y, i) for i in i_max_list]

    index_best = np.argmax([abs(y) for y in y_max_list])
    return i_max_list[index_best], y_max_list[index_best]



"""Package teneva, module anova: ANOVA decomposition in the TT-format.

This module contains the function "anova" which computes the TT-approximation
for the tensor, using given random samples.

"""
import numpy as np
import pickle
import teneva


class ANOVA:
    # TODO: add docstring and add into demo
    def __init__(self, I_trn=None, y_trn=None, order=1, seed=None, fpath=None):
        self.rand = teneva._rand(seed)

        if not order in [1, 2]:
            raise ValueError('Invalid value for ANOVA order (should be 1 or 2')
        self.order = order

        if fpath is None:
            if I_trn is None or y_trn is None:
                raise ValueError('"(I_trn, y_trn)" or "fpath" should be set')
            self.build(I_trn, y_trn)
        else:
            if I_trn is not None or y_trn is not None:
                raise ValueError('Can`t use both "(I_trn, y_trn)" and "fpath"')
            self.load(fpath)

    def __call__(self, I):
        I = np.asanyarray(I, dtype=self.dtype)

        if len(I.shape) == 1:
            return self.calc(I)
        elif len(I.shape) == 2:
            return np.array([self.calc(i) for i in I])
        else:
            raise ValueError('Invalid input multi-indices for ANOVA')

    def __getitem__(self, I):
        return self(I)

    @property
    def f1_arr(self):
        try:
            return self._f1_arr
        except AttributeError:
            pass

        f1_arr = self._f1_arr = [np.array([f1_curr[x] for x in dm])
            for dm, f1_curr in zip(self.domain, self.f1)]
        return f1_arr

    @property
    def f2_arr(self):
        try:
            return self._f2_arr
        except AttributeError:
            pass

        f2_arr = self._f2_arr = [
            np.array([f2_curr[x1, x2] for x1 in dm1 for x2 in dm2 ])
                for (dm1, dm2), f2_curr in  zip([(dm1, dm2)
                    for k1, dm1 in enumerate(self.domain[:-1])
                        for dm2 in self.domain[k1+1:]], self.f2)]
        return f2_arr

    def build(self, I_trn, y_trn):
        I_trn = np.asanyarray(I_trn)
        self.dtype = I_trn.dtype

        y_trn = np.asanyarray(y_trn, dtype=float)

        self.y_max = np.max(y_trn)
        self.y_min = np.min(y_trn)

        self.d = I_trn.shape[1]

        self.domain = []
        self.shapes = np.zeros(self.d, dtype=int)
        for k in range(self.d):
            points = np.unique(I_trn[:, k])
            self.domain.append(points)
            self.shapes[k] = len(points)

        self.build_0(I_trn, y_trn)

        if self.order >= 1:
            self.build_1(I_trn, y_trn)
        else:
            self.f1 = []

        if self.order >= 2:
            self.build_2(I_trn, y_trn)
        else:
            self.f2 = []

    def build_0(self, I_trn, y_trn):
        self.f0 = np.mean(y_trn)

    def build_1(self, I_trn, y_trn):
        self.f1 = []
        for k, dm in enumerate(self.domain):
            f1_curr = {}
            for x in dm:
                idx = I_trn[:, k] == x
                value = np.mean(y_trn[idx]) - self.f0
                f1_curr[x] = value
            self.f1.append(f1_curr)

    def build_2(self, I_trn, y_trn):
        cache = dict()
        self.f2 = []
        for k1, dm1 in enumerate(self.domain[:-1]):
            for k2, dm2 in enumerate(self.domain[k1+1:], start=k1+1):
                f2_curr = {}
                for x1 in dm1:
                    for x2 in dm2:
                        try:
                            idx1 = cache[k1, x1]
                        except KeyError:
                            idx1 = I_trn[:, k1] == x1
                            cache[k1, x1] = idx1

                        try:
                            idx2 = cache[k2, x2]
                        except KeyError:
                            idx2 = I_trn[:, k2] == x2
                            cache[k2, x2] = idx2

                        idx = idx1 & idx2

                        if idx.sum() == 0:
                            value = 0.
                        else:
                            value = np.mean(y_trn[idx]) - self.f0
                            value = value - self.f1[k1][x1] - self.f1[k2][x2]
                        f2_curr[x1, x2] = value
                self.f2.append(f2_curr)

    def calc(self, i):
        res = self.calc_0()
        if self.order >= 1:
            res += self.calc_1(i)
        if self.order >= 2:
            res += self.calc_2(i)
        return res

    def calc_0(self):
        return self.f0

    def calc_1(self, x):
        res = 0.
        for num, x1 in enumerate(x):
            res += self.f1[num][x1]
        return res

    def calc_2(self, x):
        # We allow x to be of smaller length
        res = 0.
        for i1, x1 in enumerate(x[:-1]):
            for i2, x2 in enumerate(x[i1+1:], start=i1+1):
                num = self.pair_num_to_num(i1, i2)
                res += self.f2[num][x1, x2]
        return res

    def cores(self, r=2, noise=1.E-10, only_near=False, rel_noise=None):
        if self.order < 1:
            raise ValueError('TT-cores may be constructed only if order >= 1')

        if rel_noise is not None:
            noise = rel_noise * max(abs(self.y_max), abs(self.y_min))

        cores = self.cores_1(r, noise)

        if self.order >= 2:
            cores2_many = self.cores_2(r, only_near)
            cores = teneva.add_many([cores] + cores2_many, r=r)

        return cores

    def cores_1(self, r=2, noise=1.E-10):
        cores = []

        core = noise * self.rand.normal(size=(1, self.shapes[0], r))
        core[0, :, 0] = 1.
        core[0, :, 1] = self.f1_arr[0]
        cores.append(core)

        for i in range(1, self.d-1):
            core = noise * self.rand.normal(size=(r, self.shapes[i], r))
            core[0, :, 0] = 1.
            core[1, :, 1] = 1.
            core[0, :, 1] = self.f1_arr[i]
            cores.append(core)

        core = noise * self.rand.normal(size=(r, self.shapes[self.d-1], 1))
        core[0, :, 0] = self.f1_arr[self.d-1] + self.f0
        core[1, :, 0] = 1.
        cores.append(core)

        return cores

    def cores_2(self, r=2, only_near=False):
        mats = []
        num = 0
        for i1 in range(self.d-1):
            for i2 in ([i1+1] if only_near else range(i1+1, self.d)):
                shape = (self.shapes[i1], self.shapes[i2])
                mat = self.f2_arr[num].reshape(shape, order='C')
                mats.append(mat)
                num += 1

        cores = []
        num = 0
        for i1 in range(self.d-1):
            for i2 in ([i1+1] if only_near else range(i1+1, self.d)):
                cores.append(_second_order_2_tt(mats[num], i1, i2, self.shapes))
                num += 1

        return cores

    def load(self, fpath):
        if not '.' in fpath:
            fpath += '.pickle'

        with open(fpath, 'rb') as f:
            data = pickle.load(f)

        self.order = data['order']
        self.dtype = data['dtype']
        self.y_max = data['y_max']
        self.y_min = data['y_min']
        self.domain = data['domain']
        self.shapes = data['shapes']
        self.d = data['d']
        self.f0 = data['f0']
        self.f1 = data['f1']
        self.f2 = data['f2']

    def max(self, minmax=max):
        """Get min or max value of tensor from 1th order ANOVA.

        Args:
            minmax (function): what to find (min or max).

        Returns:
            (value, np.ndarray): value of the found optimum and multi-index.

        """
        max_np = {min: np.argmin, max: np.argmax}[minmax]
        val = self.f0
        ind = [None] * self.d
        for i, fi in enumerate(self.f1):
            xx = list(fi)
            xx_max = xx[max_np([fi[x] for x in xx])]
            ind[i] = xx_max
            val += fi[xx_max]
        return val, ind

    def pair_num_to_num(self, x1, x2):
        assert x1 != x2
        if x1 > x2:
            x1, x2 = x2, x1
        return x2 - 1 + ((-3 + 2*self.d - x1)*x1) // 2

    def sample(self, xi=None, eps=1.E-10, with_square=False):
        if xi is None:
            xi = self.d - 1

        if xi > 0:
            prev_vals = self.sample(xi-1, eps, with_square)
        else:
            prev_vals = []

        dm = self.domain[xi]
        p = np.full(len(dm), self(prev_vals))

        if self.order >= 1:
            f1 = self.f1[xi]
            for i, x in enumerate(dm):
                p[i] += f1[x]

        if self.order >= 2:
            for x_prev_num, x_prev_val in enumerate(prev_vals):
                f2 = self.f2[self.pair_num_to_num(x_prev_num, xi)]
                for i, x in enumerate(dm):
                    p[i] += f2[x_prev_val, x]

        p = p**2 if with_square else np.maximum(p, 0)
        if p.sum() < eps * len(p):
            print('Warning (ANOVA): probabilities are zeros') # TODO
            p += eps

        p /= p.sum()
        p_sample = self.rand.choice(len(p), p=p)
        cur_sample = dm[p_sample]

        prev_vals.append(cur_sample)

        return prev_vals

    def save(self, fpath):
        if not '.' in fpath:
            fpath += '.pickle'

        with open(fpath, 'wb') as f:
            pickle.dump({
                'order': self.order,
                'dtype': self.dtype,
                'y_max': self.y_max,
                'y_min': self.y_min,
                'domain': self.domain,
                'shapes': self.shapes,
                'd': self.d,
                'f0': self.f0,
                'f1': self.f1,
                'f2': self.f2
            }, f, protocol=pickle.HIGHEST_PROTOCOL)


def anova(I_trn, y_trn, r=2, order=1, noise=1.E-10, seed=None, fpath=None):
    """Build TT-tensor by TT-ANOVA from the given random tensor samples.

    Args:
        I_trn (np.ndarray): multi-indices for the tensor in the form of array
            of the shape [samples, d].
        y_trn (np.ndarray): values of the tensor for multi-indices I in the
            form of array of the shape [samples].
        r (int): rank of the constructed TT-tensor.
        order (int): order of the ANOVA decomposition (may be only 1 or 2).
        noise (float): noise added to formally zero elements of TT-cores.
        seed (int): random seed. It should be an integer number or a numpy
            Generator class instance.
        fpath (str): optional path for train data (I_trn, y_trn).

    Returns:
        list: TT-tensor, which represents the TT-approximation for the tensor.

    Note:
        A class "ANOVA" that represents a wider set of methods for working with
        this decomposition is also available. See "anova.py" for more details
        (detailed documentation for this class will be prepared later). This
        function is just a wrapper for "ANOVA" class. Maybe later this class
        will be replaced by the function.

    """
    return ANOVA(I_trn, y_trn, order, seed, fpath).cores(r, noise)


def _core_one(n, r):
    return np.kron(np.ones([1, n, 1]), np.eye(r)[:, None, :])


def _second_order_2_tt(A, i, j, shapes):
    if i > j:
        j, i = i, j
        A = A.T

    U, V = teneva.matrix_skeleton(A)
    r = U.shape[1]

    core1 = U.reshape(1, U.shape[0], r)
    core2 = V.reshape(r, V.shape[1], 1)

    cores = []
    for num, n in enumerate(shapes):
        if num < i or num > j:
            cores.append(np.ones([1, n, 1]))
        if num == i:
            cores.append(core1)
        if num == j:
            cores.append(core2)
        if i < num < j:
            cores.append(_core_one(n, r))

    return cores



import numpy as np
import teneva
from time import perf_counter as tpc
import unittest


class TestCrossCross(unittest.TestCase):
    def setUp(self):
        self.d = 5
        self.n = [ 20,  18,  16,  14,  12]
        self.func = lambda I: np.sum(I, axis=1)

        self.I_tst = teneva.sample_rand(self.n, 1.E+4, seed=42)
        self.y_tst = self.func(self.I_tst)

    def test_base(self):
        r = 1
        m = None
        e = 1.E-8
        nswp = None
        dr_min = 1
        info = {}
        Y = teneva.rand(self.n, r)
        Y = teneva.cross(self.func, Y, m, e, nswp, dr_min=dr_min, dr_max=3,
            info=info)

        r1 = teneva.erank(Y)
        r2 = r + dr_min * info['nswp']
        self.assertTrue(r1 >= r2)

        Y = teneva.truncate(Y, e)

        err = teneva.accuracy_on_data(Y, self.I_tst, self.y_tst)
        self.assertLess(err, e*10)

    def test_cb(self):
        nswp = 5
        info = {}

        def cb(Y, info, opts):
            if info['nswp'] == nswp:
                # Stop the algorithm's work (just for demo!)
                return True

        Y = teneva.rand(self.n, 1)
        Y = teneva.cross(self.func, Y, nswp=100, cb=cb, info=info)

        self.assertEqual(info['nswp'], nswp)

    def test_e(self):
        r = 1
        m = None
        e = 1.E-4
        nswp = None
        dr_min = 1
        info = {}
        Y = teneva.rand(self.n, r)
        Y = teneva.cross(self.func, Y, m, e, nswp, dr_min=dr_min, dr_max=3,
            info=info)

        self.assertLess(info['e'], e)

    def test_m(self):
        r = 1
        m = 1.E+4
        e = None
        nswp = None
        dr_min = 1
        info = {}
        Y = teneva.rand(self.n, r)
        Y = teneva.cross(self.func, Y, m, e, nswp, dr_min=dr_min, dr_max=3,
            info=info)

        self.assertLess(info['m'], m+1)

    def test_m_cache_scale(self):
        nswp = 5
        cache = {}
        info = {}

        Y = teneva.rand(self.n, 1)
        Y = teneva.cross(self.func, Y, nswp=100,
            m_cache_scale=2, cache=cache, info=info)

        self.assertEqual(info['stop'], 'conv')

    def test_nswp(self):
        r = 1
        m = None
        e = None
        nswp = 4
        dr_min = 1
        info = {}
        Y = teneva.rand(self.n, r)
        Y = teneva.cross(self.func, Y, m, e, nswp, dr_min=dr_min, dr_max=3,
            info=info)

        self.assertEqual(info['nswp'], nswp)

    def test_vld(self):
        r = 1
        m = 1.E+4
        e = None
        e_vld=1.E-6
        nswp = None
        dr_min = 1
        info = {}
        I_vld = teneva.sample_lhs(self.n, 1.E+4, seed=42)
        y_vld = self.func(I_vld)
        Y = teneva.rand(self.n, r)
        Y = teneva.cross(self.func, Y, m, e, nswp, dr_min=dr_min, dr_max=3,
            info=info, I_vld=I_vld, y_vld=y_vld, e_vld=e_vld)

        err = teneva.accuracy_on_data(Y, I_vld, y_vld)
        self.assertLess(err, e_vld)

        err = teneva.accuracy_on_data(Y, self.I_tst, self.y_tst)
        self.assertLess(err, e_vld*10)


if __name__ == '__main__':
    np.random.seed(42)
    unittest.main()



import numpy as np
import teneva
from time import perf_counter as tpc
import unittest


class TestGridPoiScale(unittest.TestCase):
    def setUp(self):
        self.eps = 1.E-16

    def test_base_cheb(self):
        a = [-5., -3., -1.]
        b = [+5., +3., +1.]

        X = np.array([       # We prepare 4 spatial points:
            [-5., -3., -1.], # Point near the lower bound
            [ 0.,  0.,  0.], # Zero point
            [+5., +3., +1.], # Point near the upper bound
        ])

        Xsc = teneva.poi_scale(X, a, b, 'cheb')

        self.assertLess(abs(Xsc[0, 0] + 1.), self.eps)
        self.assertLess(abs(Xsc[0, 1] + 1.), self.eps)
        self.assertLess(abs(Xsc[0, 2] + 1.), self.eps)

        self.assertLess(abs(Xsc[1, 0] - 0.), self.eps)
        self.assertLess(abs(Xsc[1, 1] - 0.), self.eps)
        self.assertLess(abs(Xsc[1, 2] - 0.), self.eps)

        self.assertLess(abs(Xsc[2, 0] - 1.), self.eps)
        self.assertLess(abs(Xsc[2, 1] - 1.), self.eps)
        self.assertLess(abs(Xsc[2, 2] - 1.), self.eps)

    def test_base_uni(self):
        a = [-5., -3., -1.]
        b = [+5., +3., +1.]

        X = np.array([       # We prepare 4 spatial points:
            [-5., -3., -1.], # Point near the lower bound
            [ 0.,  0.,  0.], # Zero point
            [+5., +3., +1.], # Point near the upper bound
        ])

        Xsc = teneva.poi_scale(X, a, b)

        self.assertLess(abs(Xsc[0, 0] - 0.), self.eps)
        self.assertLess(abs(Xsc[0, 1] - 0.), self.eps)
        self.assertLess(abs(Xsc[0, 2] - 0.), self.eps)

        self.assertLess(abs(Xsc[1, 0] - 0.5), self.eps)
        self.assertLess(abs(Xsc[1, 1] - 0.5), self.eps)
        self.assertLess(abs(Xsc[1, 2] - 0.5), self.eps)

        self.assertLess(abs(Xsc[2, 0] - 1.), self.eps)
        self.assertLess(abs(Xsc[2, 1] - 1.), self.eps)
        self.assertLess(abs(Xsc[2, 2] - 1.), self.eps)

    def test_same_cheb(self):
        X = np.array([
            [0., 0.5, 0.7, 1.0],
            [-1., -0.9, 0.4, 0.9]])

        Xsc = teneva.poi_scale(X, -1., +1., 'cheb')

        err = np.linalg.norm(X - Xsc)
        self.assertLess(err, self.eps)

    def test_same_uni(self):
        X = np.array([
            [0., 0.5, 0.7, 1.0],
            [0.1, 0.2, 0.4, 0.8]])

        Xsc = teneva.poi_scale(X, 0., 1.)

        err = np.linalg.norm(X - Xsc)
        self.assertLess(err, self.eps)


class TestGridPoiToInd(unittest.TestCase):
    def setUp(self):
        self.eps = 1.E-14

    def test_base_cheb(self):
        a = [-5., -3., -1.]
        b = [+5., +3., +1.]
        n = [9, 8, 7]

        X = np.array([       # We prepare 4 spatial points:
            [-5., -3., -1.], # Point near the lower bound
            [ 0.,  0.,  0.], # Zero point
            [-1., +2.,  0.], # Random point
            [+5., +3., +1.], # Point near the upper bound
        ])

        I = teneva.poi_to_ind(X, a, b, n, 'cheb')

        self.assertLess(abs(I[0, 0] - 8), self.eps)
        self.assertLess(abs(I[0, 1] - 7), self.eps)
        self.assertLess(abs(I[0, 2] - 6), self.eps)

        self.assertLess(abs(I[1, 0] - 4), self.eps)
        self.assertLess(abs(I[1, 1] - 4), self.eps)
        self.assertLess(abs(I[1, 2] - 3), self.eps)

        self.assertLess(abs(I[2, 0] - 5), self.eps)
        self.assertLess(abs(I[2, 1] - 2), self.eps)
        self.assertLess(abs(I[2, 2] - 3), self.eps)

        self.assertLess(abs(I[3, 0] - 0), self.eps)
        self.assertLess(abs(I[3, 1] - 0), self.eps)
        self.assertLess(abs(I[3, 2] - 0), self.eps)

    def test_base_uni(self):
        a = [-5., -3., -1.]
        b = [+5., +3., +1.]
        n = [9, 8, 7]

        X = np.array([       # We prepare 4 spatial points:
            [-5., -3., -1.], # Point near the lower bound
            [ 0.,  0.,  0.], # Zero point
            [-1., +2.,  0.], # Random point
            [+5., +3., +1.], # Point near the upper bound
        ])

        I = teneva.poi_to_ind(X, a, b, n)

        self.assertLess(abs(I[0, 0] - 0), self.eps)
        self.assertLess(abs(I[0, 1] - 0), self.eps)
        self.assertLess(abs(I[0, 2] - 0), self.eps)

        self.assertLess(abs(I[1, 0] - 4), self.eps)
        self.assertLess(abs(I[1, 1] - 4), self.eps)
        self.assertLess(abs(I[1, 2] - 3), self.eps)

        self.assertLess(abs(I[2, 0] - 3), self.eps)
        self.assertLess(abs(I[2, 1] - 6), self.eps)
        self.assertLess(abs(I[2, 2] - 3), self.eps)

        self.assertLess(abs(I[3, 0] - 8), self.eps)
        self.assertLess(abs(I[3, 1] - 7), self.eps)
        self.assertLess(abs(I[3, 2] - 6), self.eps)

    def test_one_poi(self):
        a = [-5., -3., -1.]
        b = [+5., +3., +1.]
        n = [9, 8, 7]

        X = [-5., -3., -1.]

        I = teneva.poi_to_ind(X, a, b, n)

        self.assertEqual(len(I.shape), 1)
        self.assertEqual(I.shape[0], len(n))
        self.assertLess(abs(I[0] - 0), self.eps)
        self.assertLess(abs(I[0] - 0), self.eps)
        self.assertLess(abs(I[0] - 0), self.eps)

    def test_same_cheb(self):
        a = [-5., -3., -1.]
        b = [+5., +3., +1.]
        n = [7, 5, 3]

        X = np.array([
            [-5., -3., -1.],  # Point near the lower bound
            [ 0.,  0.,  0.],  # Zero point
            [+5., +3., +1.],  # Point near the upper bound
        ])

        I = teneva.poi_to_ind(X, a, b, n, 'cheb')
        Z = teneva.ind_to_poi(I, a, b, n, 'cheb')

        err = np.linalg.norm(X - Z)
        self.assertLess(err, self.eps)

    def test_same_uni(self):
        a = [-5., -3., -1.]
        b = [+5., +3., +1.]
        n = [7, 5, 3]

        X = np.array([
            [-5., -3., -1.],  # Point near the lower bound
            [ 0.,  0.,  0.],  # Zero point
            [+5., +3., +1.],  # Point near the upper bound
        ])

        I = teneva.poi_to_ind(X, a, b, n)
        Z = teneva.ind_to_poi(I, a, b, n)

        err = np.linalg.norm(X - Z)
        self.assertLess(err, self.eps)


if __name__ == '__main__':
    np.random.seed(42)
    unittest.main()



import numpy as np
import teneva
from time import perf_counter as tpc
import unittest


class TestTensorsConst(unittest.TestCase):
    def setUp(self):
        self.d = 5
        self.n = [10, 11, 12, 13, 14]
        self.v = 42.
        self.I_zero = [
            [0, 0, 0, 0, 0],
            [1, 2, 3, 4, 5],
            [9, 9, 9, 9, 9],
        ]
        self.i_non_zero = [5, 5, 5, 5, 5]
        self.eps = 1.E-5

    def test_base(self):
        Y = teneva.const(self.n, self.v)

        self.assertEqual(len(Y), self.d)

        for G in Y:
            self.assertEqual(G.shape[0], 1)
            self.assertEqual(G.shape[2], 1)

        Y_full = teneva.full(Y)
        e_min = abs(np.min(Y_full) - self.v)
        e_max = abs(np.max(Y_full) - self.v)
        self.assertLess(e_min, self.eps)
        self.assertLess(e_max, self.eps)

    def test_with_i_non_zero(self):
        Y = teneva.const(self.n, self.v, self.I_zero, self.i_non_zero)

        for i in self.I_zero:
            y = teneva.get(Y, i)
            e = abs(y)
            self.assertLess(e, self.eps)

        y = teneva.get(Y, self.i_non_zero)
        e = abs(y - self.v)
        self.assertLess(e, self.eps)

    def test_with_i_zero(self):
        Y = teneva.const(self.n, self.v, self.I_zero)

        for i in self.I_zero:
            y = teneva.get(Y, i)
            e = abs(y)
            self.assertLess(e, self.eps)

        Y_full = teneva.full(Y)
        mean = np.sum(Y_full) / np.sum(Y_full > 1.E-20)
        e = abs(mean - self.v)
        self.assertLess(e, self.eps)


class TestTensorsDelta(unittest.TestCase):
    def setUp(self):
        self.d = 5
        self.n = [10, 11, 12, 13, 14]
        self.v = 42.
        self.i = [5, 6, 7, 8, 9]
        self.eps = 1.E-5

    def test_base(self):
        Y = teneva.delta(self.n, self.i, self.v)

        self.assertEqual(len(Y), self.d)

        for G in Y:
            self.assertEqual(G.shape[0], 1)
            self.assertEqual(G.shape[2], 1)

        y = teneva.get(Y, self.i)
        e = abs(y - self.v)
        self.assertLess(e, self.eps)

        s = teneva.norm(Y)
        e = abs(s - self.v)
        self.assertLess(e, self.eps)

        Y_full = teneva.full(Y)

        e = abs(Y_full[tuple(self.i)] - self.v)
        self.assertLess(e, self.eps)

        s = len([y for y in Y_full.flatten() if abs(y) > 1.E-10])
        self.assertEqual(s, 1)


class TestTensorsPoly(unittest.TestCase):
    def setUp(self):
        self.d = 5
        self.n = [10, 11, 12, 13, 14]
        self.i = np.array([2, 3, 3, 4, 5])
        self.shift = np.array([2, 3, 5, 3, 8])
        self.shift_scalar = 12.
        self.scale = 5.
        self.power = 3
        self.eps = 1.E-10

    def test_base(self):
        Y = teneva.poly(self.n, self.shift, self.power, self.scale)

        self.assertEqual(len(Y), self.d)

        for G in Y[1:]:
            self.assertEqual(G.shape[0], 2)
        for G in Y[:-1]:
            self.assertEqual(G.shape[2], 2)
        for k, G in zip(self.n, Y):
            self.assertEqual(G.shape[1], k)
        self.assertEqual(Y[0].shape[0], 1)
        self.assertEqual(Y[-1].shape[2], 1)

        y_appr = teneva.get(Y, self.i)
        y_real = self.scale * np.sum((self.i + self.shift)**self.power)

        y = teneva.get(Y, self.i)
        e = abs(y_appr - y_real)
        self.assertLess(e, self.eps)

    def test_shift_scalar(self):
        Y = teneva.poly(self.n, self.shift_scalar, self.power, self.scale)

        self.assertEqual(len(Y), self.d)

        for G in Y[1:]:
            self.assertEqual(G.shape[0], 2)
        for G in Y[:-1]:
            self.assertEqual(G.shape[2], 2)
        for k, G in zip(self.n, Y):
            self.assertEqual(G.shape[1], k)
        self.assertEqual(Y[0].shape[0], 1)
        self.assertEqual(Y[-1].shape[2], 1)

        y_appr = teneva.get(Y, self.i)
        y_real = self.scale * np.sum((self.i + self.shift_scalar)**self.power)

        y = teneva.get(Y, self.i)
        e = abs(y_appr - y_real)
        self.assertLess(e, self.eps)


class TestTensorsRand(unittest.TestCase):
    def setUp(self):
        self.d = 5
        self.n = [12, 13, 14, 15, 16]
        self.r = [1, 2, 3, 4, 5, 1]
        self.r_const = 4
        self.eps = 1.E-10

    def test_base(self):
        Y = teneva.rand(self.n, self.r)

        self.assertEqual(len(Y), self.d)
        for k in range(self.d):
            self.assertEqual(Y[k].shape, (self.r[k], self.n[k], self.r[k+1]))

    def test_custom_limit(self):
        a = 0.994
        b = 0.995
        Y = teneva.rand(self.n, self.r, a=a, b=b)

        self.assertEqual(len(Y), self.d)
        for k in range(self.d):
            for g in Y[k].flatten():
                self.assertLess(a, g)
                self.assertLess(g, b)

    def test_rank_const(self):
        Y = teneva.rand(self.n, self.r_const)
        r = [1] + [self.r_const] * (self.d-1) + [1]

        self.assertEqual(len(Y), self.d)
        for k in range(self.d):
            self.assertEqual(Y[k].shape, (r[k], self.n[k], r[k+1]))


class TestTensorsRandCustom(unittest.TestCase):
    def setUp(self):
        self.d = 5
        self.n = [12, 13, 14, 15, 16]
        self.r = [1, 2, 3, 4, 5, 1]
        self.r_const = 4
        self.eps = 1.E-10

    def test_base(self):
        v = 42.
        f = lambda sz: [v]*sz
        Y = teneva.rand_custom(self.n, self.r, f)

        self.assertEqual(len(Y), self.d)
        for k in range(self.d):
            self.assertEqual(Y[k].shape, (self.r[k], self.n[k], self.r[k+1]))

        for k in range(self.d):
            for g in Y[k].flatten():
                self.assertLess(abs(v - g), 1.E-12)

    def test_rank_const(self):
        v = 42.
        f = lambda sz: [v]*sz
        Y = teneva.rand_custom(self.n, self.r_const, f)
        r = [1] + [self.r_const] * (self.d-1) + [1]

        self.assertEqual(len(Y), self.d)
        for k in range(self.d):
            self.assertEqual(Y[k].shape, (r[k], self.n[k], r[k+1]))

        for k in range(self.d):
            for g in Y[k].flatten():
                self.assertLess(abs(v - g), 1.E-12)


class TestTensorsRandNorm(unittest.TestCase):
    def setUp(self):
        self.d = 5
        self.n = [12, 13, 14, 15, 16]
        self.r = [1, 2, 3, 4, 5, 1]
        self.r_const = 4
        self.eps = 1.E-10

    def test_base(self):
        Y = teneva.rand_norm(self.n, self.r)

        self.assertEqual(len(Y), self.d)
        for k in range(self.d):
            self.assertEqual(Y[k].shape, (self.r[k], self.n[k], self.r[k+1]))

    def test_custom_limit(self):
        m = 2.
        s = 0.001
        Y = teneva.rand_norm(self.n, self.r, m=m, s=s)

        self.assertEqual(len(Y), self.d)
        for k in range(self.d):
            for g in Y[k].flatten():
                self.assertLess(m-5.*s, g)
                self.assertLess(g, m+5.*s)

    def test_rank_const(self):
        Y = teneva.rand_norm(self.n, self.r_const)
        r = [1] + [self.r_const] * (self.d-1) + [1]

        self.assertEqual(len(Y), self.d)
        for k in range(self.d):
            self.assertEqual(Y[k].shape, (r[k], self.n[k], r[k+1]))


class TestTensorsRandStab(unittest.TestCase):
    def setUp(self):
        self.d = 10000
        self.n = [10] * self.d
        self.r = 5
        self.i = [0] * self.d
        self.eps = 1.E-6

    def test_base(self):
        Y = teneva.rand_stab(self.n, self.r)
        r = [1] + [self.r] * (self.d-1) + [1]

        self.assertEqual(len(Y), self.d)
        for k in range(self.d):
            self.assertEqual(Y[k].shape, (r[k], self.n[k], r[k+1]))

        y = teneva.get(Y, self.i)
        self.assertLess(abs(y - 1.), self.eps)


if __name__ == '__main__':
    np.random.seed(42)
    unittest.main()



"""Package teneva, module tensors: various useful TT-tensors.

This module contains the collection of functions for explicit construction of
various useful TT-tensors (random tensor, delta function, polynomial
function and others).

"""
import numpy as np
import teneva


def const(n, v=1., I_zero=None, i_non_zero=None):
    """Build a TT-tensor with all values equal to the given number.

    Note that constructed TT-tensor will have all TT-ranks equal to 1.

    Args:
        n (list, np.ndarray): shape of the tensor. It should be a list or
            a np.ndarray of the length d, where d is a number of dimensions.
        v (float): all elements of the tensor will be equal to this value.
        I_zero (list, np.ndarray): optional list of lists or a np.ndarray of the
            shape samples x d, which relates to multi-indices, where tensor
            should be zero (but this ensures that the value at index
            i_non_zero, if provided, is not affected).
        i_non_zero (list, np.ndarray): optional multi-index in the form of a
            list or a np.ndarray of the shape d, which is not affected by zero
            multi-indices (I_zero).

    Returns:
        list: TT-tensor.

    """
    d = len(n)
    s = abs(v) / v if abs(v) > 1.E-16 else v
    v = abs(v)**(1./d) if abs(v) > 1.E-16 else 1.
    Y = [np.ones([1, k, 1]) * v for k in n]
    Y[-1] *= s

    if I_zero is not None:
        k = 0
        for i_zero in I_zero:
            skiped = 0
            while True:
                if i_non_zero is None or i_zero[k] != i_non_zero[k]:
                    Y[k][0, i_zero[k], 0] = 0.
                    k += 1
                    break
                else:
                    k += 1
                    skiped += 1
                    if skiped > d:
                        raise ValueError('Can not set zero items')
                    if k >= d:
                        k = 0
            if k >= d:
                k = 0

    return Y


def delta(n, i, v=1.):
    """Build a TT-tensor that is zero everywhere except for a given multi-index.

    Args:
        n (list, np.ndarray): shape of the tensor. It should be a list or
            a np.ndarray of the length d, where d is a number of dimensions.
        i (list, np.ndarray): the multi-index for nonzero element. It should
            be a list or a np.ndarray of the length d.
        v (float): the value of the tensor at the multi-index i.

    Returns:
        list: TT-tensor.

    """
    d = len(n)
    s = abs(v) / v if abs(v) > 1.E-16 else v
    v = abs(v)**(1./d) if abs(v) > 1.E-16 else 1.
    Y = [np.zeros([1, k, 1]) for k in n]
    for k in range(d):
        Y[k][0, i[k], 0] = v
    Y[-1] *= s
    return Y


def poly(n, shift=0., power=2, scale=1.):
    """Build a TT-tensor that is a polynomial like scale * (index+shift)^power.

    Args:
        n (list, np.ndarray): shape of the tensor. It should be a list or
            a np.ndarray of the length d, where d is a number of dimensions.
        shift (float, list, np.ndarray): the shift value. It should be a list or
            a np.ndarray of the length d. It may be also float value.
        power (int): the power of the polynomial.
        scale (float): the scale for the polynomial.

    Returns:
        list: TT-tensor.

    """
    d = len(n)
    Y = []

    shift = teneva.grid_prep_opt(shift, d)

    def _get(m, j):
        return (m + shift[j])**power

    for j, k in enumerate(n):
        if j == 0:
            G = np.zeros((1, k, 2))
            for m in range(k):
                G[0, m, :] = np.array([1., _get(m, j)])

        if j > 0 and j < d-1:
            G = np.zeros((2, k, 2))
            for m in range(k):
                G[:, m, :] = np.array([
                    [1., _get(m, j)],
                    [0., 1.]])

        if j == d - 1:
            G = np.zeros((2, k, 1))
            for m in range(k):
                G[:, m, 0] = np.array([_get(m, j) * scale, scale])

        Y.append(G)

    return Y


def rand(n, r, a=-1., b=1., seed=None):
    """Construct a random TT-tensor from the uniform distribution.

    Args:
        n (list, np.ndarray): shape of the tensor. It should be a list or
            a np.ndarray of the length d, where d is a number of dimensions.
        r (int, list, np.ndarray): TT-ranks of the tensor. It should be a list
            or a np.ndarray of the length d+1 with outer elements (first and
            last) equals to 1. If all inner TT-ranks are equal, it may be the
            number, which relates to the inner TT-rank.
        a (float): minimum value for random items of the TT-cores.
        b (float): maximum value for random items of the TT-cores.
        seed (int): random seed. It should be an integer number or a numpy
            Generator class instance.

    Returns:
        list: TT-tensor.

    """
    rand = teneva._rand(seed)

    def f(size):
        return rand.uniform(a, b, size=size)

    return rand_custom(n, r, f)


def rand_custom(n, r, f=np.random.randn):
    """Construct a random TT-tensor from provided distribution f.

    Args:
        n (list, np.ndarray): shape of the tensor. It should be a list or
            a np.ndarray of the length d, where d is a number of dimensions.
        r (int, list, np.ndarray): TT-ranks of the tensor. It should be a list
            or a np.ndarray of the length d+1 with outer elements (first and
            last) equals to 1. If all inner TT-ranks are equal, it may be the
            number, which relates to the inner TT-rank.
        f (function): sampling function. You can use any function that returns
            a set of random values given the required number of samples.

    Returns:
        list: TT-tensor.

    """
    n = np.asanyarray(n, dtype=int)
    d = n.size

    if isinstance(r, (int, float)):
        r = [1] + [int(r)] * (d - 1) + [1]
    r = np.asanyarray(r, dtype=int)

    ps = np.cumsum(np.concatenate(([1], n * r[0:d] * r[1:d+1])))
    ps = ps.astype(int)
    cores = np.asanyarray(f(ps[d] - 1), dtype=float)

    Y = []
    for i in range(d):
        G = cores[ps[i]-1:ps[i+1]-1]
        Y.append(G.reshape((r[i], n[i], r[i+1]), order='F'))

    return Y


def rand_norm(n, r, m=0., s=1., seed=None):
    """Construct a random TT-tensor from the normal distribution.

    Args:
        n (list, np.ndarray): shape of the tensor. It should be a list or
            a np.ndarray of the length d, where d is a number of dimensions.
        r (int, list, np.ndarray): TT-ranks of the tensor. It should be a list
            or a np.ndarray of the length d+1 with outer elements (first and
            last) equals to 1. If all inner TT-ranks are equal, it may be the
            number, which relates to the inner TT-rank.
        m (float): mean ("centre") of the distribution.
        s (float): standard deviation of the distribution (>0).
        seed (int): random seed. It should be an integer number or a numpy
            Generator class instance.

    Returns:
        list: TT-tensor.

    """
    rand = teneva._rand(seed)

    def f(size):
        return rand.normal(m, s, size=size)

    return rand_custom(n, r, f)


def rand_stab(n, r, noise=1.E-15, seed=None):
    """Construct a random TT-tensor which is stable for large dimensions.

    Args:
        n (list, np.ndarray): shape of the tensor. It should be a list or
            a np.ndarray of the length d, where d is a number of dimensions.
        r (int, list, np.ndarray): TT-ranks of the tensor. It should be a list
            or a np.ndarray of the length d+1 with outer elements (first and
            last) equals to 1. If all inner TT-ranks are equal, it may be the
            number, which relates to the inner TT-rank.
        noise (float): small noise for diagonal and off-diagonal elements.
        seed (int): random seed. It should be an integer number or a numpy
            Generator class instance.

    Returns:
        list: TT-tensor.

    """
    n = np.asanyarray(n, dtype=int)
    d = n.size

    if isinstance(r, (int, float)):
        r = [1] + [int(r)] * (d - 1) + [1]
    r = np.asanyarray(r, dtype=int)

    rand = teneva._rand(seed)

    Y = []
    for k in range(d):
        G = rand.normal(0., noise, size=(r[k], n[k], r[k+1]))
        for p in range(n[k]):
            G[:, p, :] += np.eye(r[k], r[k+1])
        Y.append(G)

    return Y
